==> bodo-2022.7-cp39-cp39-macosx_10_15_x86_64.whl: bodo/io/helpers.py <==
"""
File that contains some IO related helpers.
"""
import pyarrow as pa
from mpi4py import MPI
from numba.core import types
from numba.core.imputils import lower_constant
from numba.extending import NativeValue, box, models, register_model, typeof_impl, unbox
import bodo
from bodo.hiframes.datetime_date_ext import datetime_date_array_type, datetime_date_type
from bodo.hiframes.pd_categorical_ext import CategoricalArrayType, PDCategoricalDtype
from bodo.libs.array_item_arr_ext import ArrayItemArrayType
from bodo.libs.binary_arr_ext import binary_array_type, bytes_type
from bodo.libs.bool_arr_ext import boolean_array
from bodo.libs.decimal_arr_ext import DecimalArrayType
from bodo.libs.dict_arr_ext import dict_str_arr_type
from bodo.libs.int_arr_ext import IntegerArrayType
from bodo.libs.str_arr_ext import string_array_type
from bodo.libs.str_ext import string_type
from bodo.libs.struct_arr_ext import StructArrayType
from bodo.utils.typing import BodoError


class PyArrowTableSchemaType(types.Opaque):

    def __init__(self):
        super(PyArrowTableSchemaType, self).__init__(name=
            'PyArrowTableSchemaType')


pyarrow_table_schema_type = PyArrowTableSchemaType()
types.pyarrow_table_schema_type = pyarrow_table_schema_type
register_model(PyArrowTableSchemaType)(models.OpaqueModel)


@unbox(PyArrowTableSchemaType)
def unbox_pyarrow_table_schema_type(typ, val, c):
    c.pyapi.incref(val)
    return NativeValue(val)


@box(PyArrowTableSchemaType)
def box_pyarrow_table_schema_type(typ, val, c):
    c.pyapi.incref(val)
    return val


@typeof_impl.register(pa.lib.Schema)
def typeof_pyarrow_table_schema(val, c):
    return pyarrow_table_schema_type


@lower_constant(PyArrowTableSchemaType)
def lower_pyarrow_table_schema(context, builder, ty, pyval):
    nbmq__mkj = context.get_python_api(builder)
    return nbmq__mkj.unserialize(nbmq__mkj.serialize_object(pyval))


def is_nullable(typ):
    return bodo.utils.utils.is_array_typ(typ, False) and (not isinstance(
        typ, types.Array) and not isinstance(typ, bodo.DatetimeArrayType))


def pa_schema_unify_reduction(schema_a, schema_b, unused):
    return pa.unify_schemas([schema_a, schema_b])


pa_schema_unify_mpi_op = MPI.Op.Create(pa_schema_unify_reduction, commute=True)
use_nullable_int_arr = True
_pyarrow_numba_type_map = {pa.bool_(): types.bool_, pa.int8(): types.int8,
    pa.int16(): types.int16, pa.int32(): types.int32, pa.int64(): types.
    int64, pa.uint8(): types.uint8, pa.uint16(): types.uint16, pa.uint32():
    types.uint32, pa.uint64(): types.uint64, pa.float32(): types.float32,
    pa.float64(): types.float64, pa.string(): string_type, pa.large_string(
    ): string_type, pa.binary(): bytes_type, pa.date32():
    datetime_date_type, pa.date64(): types.NPDatetime('ns'), pa.null():
    string_type}


def get_arrow_timestamp_type(pa_ts_typ):
    umpp__yzqca = 'ns', 'us', 'ms', 's'
    if pa_ts_typ.unit not in umpp__yzqca:
        return types.Array(bodo.datetime64ns, 1, 'C'), False
    elif pa_ts_typ.tz is not None:
        eli__vwz = pa_ts_typ.to_pandas_dtype().tz
        tzg__jbwzb = bodo.libs.pd_datetime_arr_ext.get_pytz_type_info(eli__vwz)
        return bodo.DatetimeArrayType(tzg__jbwzb), True
    else:
        return types.Array(bodo.datetime64ns, 1, 'C'), True


def _get_numba_typ_from_pa_typ(pa_typ: pa.Field, is_index,
    nullable_from_metadata, category_info, str_as_dict=False):
    if isinstance(pa_typ.type, pa.ListType):
        tokb__tszz, blyv__kggps = _get_numba_typ_from_pa_typ(pa_typ.type.
            value_field, is_index, nullable_from_metadata, category_info)
        return ArrayItemArrayType(tokb__tszz), blyv__kggps
    if isinstance(pa_typ.type, pa.StructType):
        ejo__adnph = []
        eabh__gyhfl = []
        blyv__kggps = True
        for wcxp__nxmey in pa_typ.flatten():
            eabh__gyhfl.append(wcxp__nxmey.name.split('.')[-1])
            xlu__nvyxs, tedy__vrrew = _get_numba_typ_from_pa_typ(wcxp__nxmey,
                is_index, nullable_from_metadata, category_info)
            ejo__adnph.append(xlu__nvyxs)
            blyv__kggps = blyv__kggps and tedy__vrrew
        return StructArrayType(tuple(ejo__adnph), tuple(eabh__gyhfl)
            ), blyv__kggps
    if isinstance(pa_typ.type, pa.Decimal128Type):
        return DecimalArrayType(pa_typ.type.precision, pa_typ.type.scale), True
    if str_as_dict:
        if pa_typ.type != pa.string():
            raise BodoError(
                f'Read as dictionary used for non-string column {pa_typ}')
        return dict_str_arr_type, True
    if isinstance(pa_typ.type, pa.DictionaryType):
        if pa_typ.type.value_type != pa.string():
            raise BodoError(
                f'Parquet Categorical data type should be string, not {pa_typ.type.value_type}'
                )
        rjqyy__mbh = _pyarrow_numba_type_map[pa_typ.type.index_type]
        vshhn__peiq = PDCategoricalDtype(category_info[pa_typ.name], bodo.
            string_type, pa_typ.type.ordered, int_type=rjqyy__mbh)
        return CategoricalArrayType(vshhn__peiq), True
    if isinstance(pa_typ.type, pa.lib.TimestampType):
        return get_arrow_timestamp_type(pa_typ.type)
    elif pa_typ.type in _pyarrow_numba_type_map:
        sns__gjul = _pyarrow_numba_type_map[pa_typ.type]
        blyv__kggps = True
    else:
        raise BodoError('Arrow data type {} not supported yet'.format(
            pa_typ.type))
    if sns__gjul == datetime_date_type:
        return datetime_date_array_type, blyv__kggps
    if sns__gjul == bytes_type:
        return binary_array_type, blyv__kggps
    tokb__tszz = (string_array_type if sns__gjul == string_type else types.
        Array(sns__gjul, 1, 'C'))
    if sns__gjul == types.bool_:
        tokb__tszz = boolean_array
    qmjw__znuhy = (use_nullable_int_arr if nullable_from_metadata is None else
        nullable_from_metadata)
    if qmjw__znuhy and not is_index and isinstance(sns__gjul, types.Integer
        ) and pa_typ.nullable:
        tokb__tszz = IntegerArrayType(sns__gjul)
    return tokb__tszz, blyv__kggps

==> bodo-2022.7.2-cp39-cp39-macosx_10_15_x86_64.whl: bodo/io/helpers.py <==
"""
File that contains some IO related helpers.
"""
import pyarrow as pa
from mpi4py import MPI
from numba.core import types
from numba.core.imputils import lower_constant
from numba.extending import NativeValue, box, models, register_model, typeof_impl, unbox
import bodo
from bodo.hiframes.datetime_date_ext import datetime_date_array_type, datetime_date_type
from bodo.hiframes.pd_categorical_ext import CategoricalArrayType, PDCategoricalDtype
from bodo.libs.array_item_arr_ext import ArrayItemArrayType
from bodo.libs.binary_arr_ext import binary_array_type, bytes_type
from bodo.libs.bool_arr_ext import boolean_array
from bodo.libs.decimal_arr_ext import DecimalArrayType
from bodo.libs.dict_arr_ext import dict_str_arr_type
from bodo.libs.int_arr_ext import IntegerArrayType
from bodo.libs.str_arr_ext import string_array_type
from bodo.libs.str_ext import string_type
from bodo.libs.struct_arr_ext import StructArrayType
from bodo.utils.typing import BodoError


class PyArrowTableSchemaType(types.Opaque):

    def __init__(self):
        super(PyArrowTableSchemaType, self).__init__(name=
            'PyArrowTableSchemaType')


pyarrow_table_schema_type = PyArrowTableSchemaType()
types.pyarrow_table_schema_type = pyarrow_table_schema_type
register_model(PyArrowTableSchemaType)(models.OpaqueModel)


@unbox(PyArrowTableSchemaType)
def unbox_pyarrow_table_schema_type(typ, val, c):
    c.pyapi.incref(val)
    return NativeValue(val)


@box(PyArrowTableSchemaType)
def box_pyarrow_table_schema_type(typ, val, c):
    c.pyapi.incref(val)
    return val


@typeof_impl.register(pa.lib.Schema)
def typeof_pyarrow_table_schema(val, c):
    return pyarrow_table_schema_type


@lower_constant(PyArrowTableSchemaType)
def lower_pyarrow_table_schema(context, builder, ty, pyval):
    eqf__gninh = context.get_python_api(builder)
    return eqf__gninh.unserialize(eqf__gninh.serialize_object(pyval))


def is_nullable(typ):
    return bodo.utils.utils.is_array_typ(typ, False) and (not isinstance(
        typ, types.Array) and not isinstance(typ, bodo.DatetimeArrayType))


def pa_schema_unify_reduction(schema_a, schema_b, unused):
    return pa.unify_schemas([schema_a, schema_b])


pa_schema_unify_mpi_op = MPI.Op.Create(pa_schema_unify_reduction, commute=True)
use_nullable_int_arr = True
_pyarrow_numba_type_map = {pa.bool_(): types.bool_, pa.int8(): types.int8,
    pa.int16(): types.int16, pa.int32(): types.int32, pa.int64(): types.
    int64, pa.uint8(): types.uint8, pa.uint16(): types.uint16, pa.uint32():
    types.uint32, pa.uint64(): types.uint64, pa.float32(): types.float32,
    pa.float64(): types.float64, pa.string(): string_type, pa.large_string(
    ): string_type, pa.binary(): bytes_type, pa.date32():
    datetime_date_type, pa.date64(): types.NPDatetime('ns'), pa.null():
    string_type}


def get_arrow_timestamp_type(pa_ts_typ):
    mxa__xhm = 'ns', 'us', 'ms', 's'
    if pa_ts_typ.unit not in mxa__xhm:
        return types.Array(bodo.datetime64ns, 1, 'C'), False
    elif pa_ts_typ.tz is not None:
        iiykb__dvok = pa_ts_typ.to_pandas_dtype().tz
        pbc__vau = bodo.libs.pd_datetime_arr_ext.get_pytz_type_info(iiykb__dvok
            )
        return bodo.DatetimeArrayType(pbc__vau), True
    else:
        return types.Array(bodo.datetime64ns, 1, 'C'), True


def _get_numba_typ_from_pa_typ(pa_typ: pa.Field, is_index,
    nullable_from_metadata, category_info, str_as_dict=False):
    if isinstance(pa_typ.type, pa.ListType):
        nevl__fsf, szkdn__ogocn = _get_numba_typ_from_pa_typ(pa_typ.type.
            value_field, is_index, nullable_from_metadata, category_info)
        return ArrayItemArrayType(nevl__fsf), szkdn__ogocn
    if isinstance(pa_typ.type, pa.StructType):
        pvv__pdt = []
        cpe__brym = []
        szkdn__ogocn = True
        for yoey__kdlw in pa_typ.flatten():
            cpe__brym.append(yoey__kdlw.name.split('.')[-1])
            opbw__utb, thdc__his = _get_numba_typ_from_pa_typ(yoey__kdlw,
                is_index, nullable_from_metadata, category_info)
            pvv__pdt.append(opbw__utb)
            szkdn__ogocn = szkdn__ogocn and thdc__his
        return StructArrayType(tuple(pvv__pdt), tuple(cpe__brym)), szkdn__ogocn
    if isinstance(pa_typ.type, pa.Decimal128Type):
        return DecimalArrayType(pa_typ.type.precision, pa_typ.type.scale), True
    if str_as_dict:
        if pa_typ.type != pa.string():
            raise BodoError(
                f'Read as dictionary used for non-string column {pa_typ}')
        return dict_str_arr_type, True
    if isinstance(pa_typ.type, pa.DictionaryType):
        if pa_typ.type.value_type != pa.string():
            raise BodoError(
                f'Parquet Categorical data type should be string, not {pa_typ.type.value_type}'
                )
        dif__zyn = _pyarrow_numba_type_map[pa_typ.type.index_type]
        dukwk__gmpn = PDCategoricalDtype(category_info[pa_typ.name], bodo.
            string_type, pa_typ.type.ordered, int_type=dif__zyn)
        return CategoricalArrayType(dukwk__gmpn), True
    if isinstance(pa_typ.type, pa.lib.TimestampType):
        return get_arrow_timestamp_type(pa_typ.type)
    elif pa_typ.type in _pyarrow_numba_type_map:
        aueee__fyswk = _pyarrow_numba_type_map[pa_typ.type]
        szkdn__ogocn = True
    else:
        raise BodoError('Arrow data type {} not supported yet'.format(
            pa_typ.type))
    if aueee__fyswk == datetime_date_type:
        return datetime_date_array_type, szkdn__ogocn
    if aueee__fyswk == bytes_type:
        return binary_array_type, szkdn__ogocn
    nevl__fsf = (string_array_type if aueee__fyswk == string_type else
        types.Array(aueee__fyswk, 1, 'C'))
    if aueee__fyswk == types.bool_:
        nevl__fsf = boolean_array
    zolqw__faawb = (use_nullable_int_arr if nullable_from_metadata is None else
        nullable_from_metadata)
    if zolqw__faawb and not is_index and isinstance(aueee__fyswk, types.Integer
        ) and pa_typ.nullable:
        nevl__fsf = IntegerArrayType(aueee__fyswk)
    return nevl__fsf, szkdn__ogocn
